using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;

using Syntriax.Engine.Core;

namespace Syntriax.Engine.Physics2D.Primitives;

public record Line(Vector2D From, Vector2D To)
{
    public Line Reversed => new(To, From);
    public Vector2D Direction => From.FromTo(To).Normalize();
    public float Length => From.FromTo(To).Length();
    public float LengthSquared => From.FromTo(To).LengthSquared();

    public static LineEquation GetLineEquation(Line line)
    {
        Vector2D slopeVector = line.From.FromTo(line.To);
        float slope = slopeVector.Y / slopeVector.X;

        float yOffset = line.From.Y - (slope * line.From.X);

        return new LineEquation(slope, yOffset);
    }

    public static bool Intersects(Line line, Vector2D point)
        => LineEquation.Resolve(GetLineEquation(line), point.X).ApproximatelyEquals(point.Y);

    public static float GetT(Line line, Vector2D point)
    {
        float fromX = MathF.Abs(line.From.X);
        float toX = MathF.Abs(line.To.X);
        float pointX = MathF.Abs(point.X);

        float min = MathF.Min(fromX, toX);
        float max = MathF.Max(fromX, toX) - min;

        pointX -= min;

        float t = pointX / max;

        // FIXME
        // I don't even know, apparently whatever I wrote up there doesn't take into account of the direction of the line
        // Which... I can see how, but I am also not sure how I can make it take into account. Or actually I'm for some reason
        // too unmotivated to find a solution. Future me, find a better way if possible, please.
        if (!Lerp(line, t).ApproximatelyEquals(point))
            return 1f - t;
        return t;
    }

    public static bool Exist(Line line, List<Vector2D> vertices)
    {
        for (int i = 0; i < vertices.Count - 1; i++)
        {
            Vector2D vertexCurrent = vertices[i];
            Vector2D vertexNext = vertices[i];
            if (line.From == vertexCurrent && line.To == vertexNext) return true;
            if (line.From == vertexNext && line.To == vertexCurrent) return true;
        }

        Vector2D vertexFirst = vertices[0];
        Vector2D vertexLast = vertices[^1];
        if (line.From == vertexFirst && line.To == vertexLast) return true;
        if (line.From == vertexLast && line.To == vertexFirst) return true;
        return false;
    }

    public static float IntersectionParameterT(Line left, Line right)
    {
        float numerator = (left.From.X - right.From.X) * (right.From.Y - right.To.Y) - (left.From.Y - right.From.Y) * (right.From.X - right.To.X);
        float denominator = (left.From.X - left.To.X) * (right.From.Y - right.To.Y) - (left.From.Y - left.To.Y) * (right.From.X - right.To.X);

        // Lines are parallel
        if (denominator == 0)
            return float.NaN;

        return numerator / denominator;
    }

    public static Vector2D Lerp(Line line, float t)
        => new Vector2D(
            line.From.X + (line.To.X - line.From.X) * t,
            line.From.Y + (line.To.Y - line.From.Y) * t
        );

    public static Vector2D ClosestPointTo(Line line, Vector2D point)
    {
        // Convert edge points to vectors
        var edgeVector = new Vector2D(line.To.X - line.From.X, line.To.Y - line.From.Y);
        var pointVector = new Vector2D(point.X - line.From.X, point.Y - line.From.Y);

        // Calculate the projection of pointVector onto edgeVector
        float t = (pointVector.X * edgeVector.X + pointVector.Y * edgeVector.Y) / (edgeVector.X * edgeVector.X + edgeVector.Y * edgeVector.Y);

        // Clamp t to the range [0, 1] to ensure the closest point is on the edge
        t = MathF.Max(0, MathF.Min(1, t));

        // Calculate the closest point on the edge
        float closestX = line.From.X + t * edgeVector.X;
        float closestY = line.From.Y + t * edgeVector.Y;

        return new Vector2D((float)closestX, (float)closestY);
    }

    public static Vector2D IntersectionPoint(Line left, Line right)
        => Vector2D.Lerp(left.From, left.To, IntersectionParameterT(left, right));

    public static bool Intersects(Line left, Line right)
    {
        int o1 = PhysicsMath.Orientation(left.From, left.To, right.From);
        int o2 = PhysicsMath.Orientation(left.From, left.To, right.To);
        int o3 = PhysicsMath.Orientation(right.From, right.To, left.From);
        int o4 = PhysicsMath.Orientation(right.From, right.To, left.To);

        if (o1 != o2 && o3 != o4)
            return true;

        if (o1 == 0 && PhysicsMath.OnSegment(left.From, right.From, left.To)) return true;
        if (o2 == 0 && PhysicsMath.OnSegment(left.From, right.To, left.To)) return true;
        if (o3 == 0 && PhysicsMath.OnSegment(right.From, left.From, right.To)) return true;
        if (o4 == 0 && PhysicsMath.OnSegment(right.From, left.To, right.To)) return true;

        return false;
    }

    public static bool Intersects(Line left, Line right, [NotNullWhen(returnValue: true)] out Vector2D? point)
    {
        point = null;

        bool result = Intersects(left, right);

        if (result)
            point = IntersectionPoint(left, right);

        return result;
    }

    public static bool ApproximatelyEquals(Line left, Line right)
        => left.From.ApproximatelyEquals(right.From) && left.To.ApproximatelyEquals(right.To);
}

public static class LineExtensions
{
    public static bool ApproximatelyEquals(this Line left, Line right) => Line.ApproximatelyEquals(left, right);
}
